/*
 * Hydrogen
 * Copyright(c) 2002-2003 by Alex >Comix< Cominu [comix@users.sourceforge.net]
 *
 * http://hydrogen.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: Hydrogen.cpp,v 1.3 2003/12/11 09:05:07 comix Exp $
 *
 */
#include "config.h"

#include <pthread.h>
#include <deque>
#include <sys/time.h>
#include <iostream>
#include <time.h>
#include <math.h>
using std::cout;
using std::cerr;
using std::endl;

#include "Sample.h"
#include "LocalFileMng.h"
#include "Hydrogen.h"

#include "JackDriver.h"
#include "OssDriver.h"
#include "NullDriver.h"
#include "DiskWriterDriver.h"
#include "AlsaMidiDriver.h"
#include "LadspaFX.h"


// GLOBALS
/** Mutex for syncronized access to song object */
pthread_mutex_t m_engineLock_mutex;


/** Hydrogen class instance (used for log) */
Hydrogen* hydrogenInstance = NULL;


/** Audio Driver */
GenericDriver *m_pAudioDriver = NULL;

/** Song Note FIFO */
std::deque<Note*> songNoteQueue;


// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ
/** Midi Driver */
AlsaMidiDriver *midiDriver = NULL;
#endif

unsigned maxNotes;

/** Midi Note FIFO */
std::deque<Note*> midiNoteQueue;

/** Playing note FIFO */
std::vector<Note*> playingNotesQueue;

unsigned long m_nTotalFrames = 0;

uint tickSize;

/** Current song */
Song *m_pSong;

/** Current pattern */
Pattern* m_pPattern;

/** Next pattern (used only in Live mode) */
Pattern* m_pNextPattern;

int m_nPatternPos;

/** Metronome Instrument */
Instrument *metronomeInstrument = NULL;

/** Master peak (left channel) */
float masterPeak_L = 0.0f;

/** Master peak (right channel) */
float masterPeak_R = 0.0f;

// questi buffer vengono usati nella funzione process
uint audioEngine_buffer_size = 0;
float *main_buffer_L = NULL;
float *main_buffer_R = NULL;
float *track_buffers_L[MAX_TRACKS];
float *track_buffers_R[MAX_TRACKS];
bool use_track_outs = false;
bool use_default_outs = false;

/** static reference of Hydrogen class */
Hydrogen* Hydrogen::instance = NULL;

/** number of playing notes */
uint playingNotes = 0;

/** Audio engine state */
int audioEngine_state = UNINITIALIZED;

std::vector<EngineListener*> audioEngine_listeners;

/** time used in process function */
float processTime = 0.0f;

/** max ms usable in process with no xrun */
float maxProcessTime = 0.0f;

/** Delay FX */
#define MAX_DELAY  44100 * 2
float delayRingBuffer_L[MAX_DELAY];	// 1sec delay
float delayRingBuffer_R[MAX_DELAY];	// 1sec delay
float *delayBuffer_L = NULL;
float *delayBuffer_R = NULL;
float delayBufferPeak_L = 0.0f;
float delayBufferPeak_R = 0.0f;
int delayStartPos = 0;


// LADSPA
LadspaFX* m_pLadspaFX = NULL;
float* m_pFxBuffer_L = NULL;
float* m_pFxBuffer_R = NULL;



int m_nPatternStartTick = -1;
uint m_nPatternTickPosition = 0;

// used in findPatternInTick
int findPatternInTick_lastTick = 0;
int findPatternInTick_lastTotal = 0;



// PROTOTYPES
void	audioEngine_init();
void	audioEngine_destroy();
int	audioEngine_start(bool bLockEngine = false, uint nTotalFrames = 0);
void	audioEngine_stop(bool bLockEngine = false);
void	audioEngine_setSong(Song *newSong);
void	audioEngine_removeSong();
void	audioEngine_noteOn(Note *note);
void	audioEngine_noteOff(Note *note);
int	audioEngine_process(uint32_t nframes, void *arg);
inline void audioEngine_clearNoteQueue();
inline void audioEngine_process_checkBPMChanged();
inline void audioEngine_process_playPattern(unsigned long nframes);
inline void audioEngine_process_playMidi(unsigned long nframes);
inline void audioEngine_process_updateInfo();

inline uint audioEngine_renderNote(Note* note, int nFrames, uint bufferSize);
inline int audioEngine_updateNoteQueue(uint nFrames);

inline int findPatternInTick( int tick, bool loopMode, int *patternStartTick );

void audioEngine_seek( unsigned long nFrames );

void audioEngine_restartAudioDrivers();
void audioEngine_startAudioDrivers();
void audioEngine_stopAudioDrivers();


//----------------------------------------------------------------------------
inline unsigned long currentTime() {
	struct timeval now;
	gettimeofday(&now, NULL);
	return now.tv_sec * 1000 + now.tv_usec / 1000;
}



//----------------------------------------------------------------------------
inline timeval currentTime2() {
	struct timeval now;
	gettimeofday(&now, NULL);
	return now;
}



//----------------------------------------------------------------------------
inline int randomValue( int max ) {
	return rand() % max;
}


//----------------------------------------------------------------------------
inline void audioEngine_lock() {
	pthread_mutex_lock(&m_engineLock_mutex);
}


//----------------------------------------------------------------------------
inline void audioEngine_unlock() {
	pthread_mutex_unlock(&m_engineLock_mutex);
}



//----------------------------------------------------------------------------
void audioEngine_raiseError( unsigned nErrorCode ) {
	// raise error event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->engineError( nErrorCode );
	}
}



//----------------------------------------------------------------------------
void updateTickSize() {
	hydrogenInstance->infoLog("UpdateTickSize");
	float bpm = m_pSong->getBpm();
	float resolution = m_pSong->getResolution();
	float sampleRate = (float)m_pAudioDriver->getSampleRate();
	tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );
}





//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_init() {
	hydrogenInstance->infoLog("Hydrogen audio engine init");

	// check current state
	if (audioEngine_state != UNINITIALIZED) {
		hydrogenInstance->errorLog("audioEngine_init: Error the audio engine is not in UNINITIALIZED state");
		audioEngine_unlock();
		return;
	}

	m_pSong = NULL;
	m_pPattern = NULL;
	m_pNextPattern = NULL;
	m_nPatternPos = -1;
	m_nPatternTickPosition = 0;
	metronomeInstrument = NULL;
	m_pAudioDriver = NULL;

	main_buffer_L = NULL;
	main_buffer_R = NULL;
	delayBuffer_L = NULL;
	delayBuffer_R = NULL;

	m_pFxBuffer_L = NULL;
	m_pFxBuffer_R = NULL;

	for (unsigned int n=0; n < MAX_TRACKS; ++n) {
		track_buffers_L[n] = NULL;
		track_buffers_R[n] = NULL;
	}

	memset( delayRingBuffer_L, 0, MAX_DELAY * sizeof(float));
	memset( delayRingBuffer_R, 0, MAX_DELAY * sizeof(float));

	PreferencesMng *preferences = PreferencesMng::getInstance();

	maxNotes = preferences->getMaxNotes();
	use_track_outs = preferences->isJackTrackOuts();
	use_default_outs = preferences->isJackConnectDefaults();


	pthread_mutex_init(&m_engineLock_mutex, NULL);

	srand(time(NULL));

	// Create metronome instrument
	metronomeInstrument = Instrument::load( string(DATA_DIR) + "/click.wav" );
	metronomeInstrument->setVolume( 0.0 );

	// set initial metronome volume
	if ( preferences->isMetronomeEnabled() ) {
		metronomeInstrument->setVolume( preferences->getMetronomeVolume() );
	}
	else {
		metronomeInstrument->setVolume( 0.0 );
	}

	// Change the current audio engine state
	audioEngine_state = INITIALIZED;
	hydrogenInstance->infoLog( "INITIALIZED" );

	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged( INITIALIZED );
	}
}





//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_destroy() {
	audioEngine_lock();
	hydrogenInstance->infoLog("Hydrogen audio engine destroy");

	// check current state
	if (audioEngine_state != INITIALIZED) {
		hydrogenInstance->errorLog("audioEngine_destroy: Error the audio engine is not in INITIALIZED state");
		return;
	}

	// change the current audio engine state
	audioEngine_state = UNINITIALIZED;
	hydrogenInstance->infoLog( "UNINITIALIZED" );

	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(UNINITIALIZED);
	}


	// delete all copied notes in the song notes queue
	for (uint i = 0; i < songNoteQueue.size(); i++) {
		Note *note = songNoteQueue[i];
		delete note;
		note = NULL;
	}
	songNoteQueue.clear();

	// delete all copied notes in the playing notes queue
	for (uint i = 0; i < playingNotesQueue.size(); i++) {
		Note *note = playingNotesQueue[i];
		delete note;
		note = NULL;
	}
	playingNotesQueue.clear();

	// delete all copied notes in the midi notes queue
	for (uint i = 0; i < midiNoteQueue.size(); i++) {
		Note *note = midiNoteQueue[i];
		delete note;
		note = NULL;
	}
	midiNoteQueue.clear();

	hydrogenInstance->infoLog( "audioEngine_destroy: all notes buffers cleared" );

	hydrogenInstance->infoLog("Destroying delayBuffer_L");
	delete[] delayBuffer_L;
	delayBuffer_L = NULL;

	hydrogenInstance->infoLog("Destroying delayBuffer_R");
	delete[] delayBuffer_R;
	delayBuffer_R = NULL;

	delete metronomeInstrument;
	metronomeInstrument = NULL;

	delete m_pFxBuffer_L;
	m_pFxBuffer_L = NULL;

	delete m_pFxBuffer_R;
	m_pFxBuffer_R = NULL;

	delete m_pLadspaFX;

	audioEngine_unlock();
}





//----------------------------------------------------------------------------
/**
 * Start playing
 * return 0 = OK
 * return -1 = NULL Audio Driver
 * return -2 = Driver connect() error
 */
int audioEngine_start(bool bLockEngine, uint nTotalFrames) {
	if (bLockEngine) {
		audioEngine_lock();
	}

	hydrogenInstance->infoLog("Audio engine starting...");

	// check current state
	if (audioEngine_state != READY) {
		hydrogenInstance->errorLog("audioEngine_start: Error the audio engine is not in READY state");
		if (bLockEngine) {
			audioEngine_unlock();
		}
		return 0;	// FIXME!!
	}


	PreferencesMng *preferencesMng = PreferencesMng::getInstance();
	masterPeak_L = 0.0f;
	masterPeak_R = 0.0f;
	delayBufferPeak_L = 0.0f;
	delayBufferPeak_R = 0.0f;
	m_nTotalFrames = nTotalFrames;	// reset total frames
	m_nPatternPos = -1;
	m_nPatternStartTick = -1;
	m_nPatternTickPosition = 0;

	// reset
	findPatternInTick_lastTotal = 0;
	findPatternInTick_lastTick = 0;

	// prepare the tickSize for this song
	updateTickSize();

	maxNotes = preferencesMng->getMaxNotes();

	// set initial metronome volume
	if ( preferencesMng->isMetronomeEnabled() ) {
		metronomeInstrument->setVolume( preferencesMng->getMetronomeVolume() );
	}
	else {
		metronomeInstrument->setVolume( 0.0 );
	}

	// change the current audio engine state
	audioEngine_state = PLAYING;
	hydrogenInstance->infoLog("PLAYING");
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(PLAYING);
	}

	if (bLockEngine) {
		audioEngine_unlock();
	}
	return 0; // per ora restituisco sempre OK
}





//----------------------------------------------------------------------------
/**
 * Stop the audio engine
 */
void audioEngine_stop(bool bLockEngine) {
	if (bLockEngine) {
		audioEngine_lock();
	}
	hydrogenInstance->infoLog("Stopping audio engine...");

	// check current state
	if (audioEngine_state != PLAYING) {
		hydrogenInstance->errorLog("audioEngine_stop: Error the audio engine is not in PLAYING state");
		cout << "state = " << audioEngine_state << endl;
		if (bLockEngine) {
			audioEngine_unlock();
		}
		return;
	}


	// change the current audio engine state
	audioEngine_state = READY;
	hydrogenInstance->infoLog("READY");
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(READY);
	}

	masterPeak_L = 0.0f;
	masterPeak_R = 0.0f;
	delayBufferPeak_L = 0.0f;
	delayBufferPeak_R = 0.0f;
	m_nPatternTickPosition = 0;
	m_nPatternStartTick = -1;

	// delete all copied notes in the song notes queue
	for (uint i = 0; i < songNoteQueue.size(); i++) {
		Note *note = songNoteQueue[i];
		if (note) {
			delete note;
			note = NULL;
		}
	}
	songNoteQueue.clear();

/*
	// delete all copied notes in the playing notes queue
	for (uint i = 0; i < playingNotesQueue.size(); i++) {
		Note *note = playingNotesQueue[i];
		if (note) {
			delete note;
			note = NULL;
		}
	}
	playingNotesQueue.clear();
*/

	// delete all copied notes in the midi notes queue
	for (uint i = 0; i < midiNoteQueue.size(); i++) {
		Note *note = midiNoteQueue[i];
		if (note) {
			delete note;
			note = NULL;
		}
	}
	midiNoteQueue.clear();

	if (bLockEngine) {
		audioEngine_unlock();
	}
}




//----------------------------------------------------------------------------
/**
 *
 */
inline void audioEngine_process_checkBPMChanged()
{
	uint newTickSize = (uint)((float)m_pAudioDriver->getSampleRate() * 60.0 /  m_pSong->getBpm() / m_pSong->getResolution());

	if (newTickSize != tickSize) {
		// update tickSize
		tickSize = newTickSize;
		hydrogenInstance->infoLog("[audioEngine_process_checkBPMChanged()] tickSize changed");
		audioEngine_clearNoteQueue();
	}
}





//----------------------------------------------------------------------------
/**
 *
 */
inline void audioEngine_process_playPattern( unsigned long nframes )
{
	if ( audioEngine_state == PLAYING ) {
		// leggo da songNoteQueue
		while (songNoteQueue.size() > 0) {
			Note *note = songNoteQueue[0];

			// verifico se la nota rientra in questo ciclo
			uint noteStartInFrames = note->getPosition() * tickSize;

			// m_nTotalFrames <= NotePos < m_nTotalFrames + bufferSize
			bool isNoteStart = ( ( noteStartInFrames >= m_nTotalFrames ) && ( noteStartInFrames < ( m_nTotalFrames + nframes ) ) );
			bool isOldNote = noteStartInFrames < m_nTotalFrames;
			if ( isNoteStart || isOldNote ) {
				playingNotesQueue.push_back( note );		// aggiungo la nota alla lista di note da eseguire
				songNoteQueue.pop_front();			// rimuovo la nota dalla lista di note

				// exclude notes
				Instrument *instr = note->getInstrument();
				if ( instr->excludeVect.size() != 0 ) {
					for (uint i = 0; i < instr->excludeVect.size(); i++) {
						Instrument *pExcluded = instr->excludeVect[ i ];
						// remove all notes listed in excludeVect
						for ( uint j = 0; j < playingNotesQueue.size(); j++ ) {	// delete older note
							Note *oldNote = playingNotesQueue[ j ];

							if ( oldNote->getInstrument() == pExcluded ) {
//								hydrogenInstance->infoLog("Delete excluded note");
								playingNotesQueue.erase( playingNotesQueue.begin() + j );
								delete oldNote;
							}
						}
					}
				}

				// raise noteOn event
				for ( uint nList = 0; nList < audioEngine_listeners.size(); nList++ ) {
					audioEngine_listeners[nList]->noteOn( note->copy() );	// the note must be deleted!
				}
				continue;
			}
			else {
				// la nota non andra' in esecuzione
				break;
			}
		}
	}

}




//----------------------------------------------------------------------------
/**
 *
 */
inline void audioEngine_process_playMidi( unsigned long nframes )
{
	// leggo da midiNoteQueue
	while ( midiNoteQueue.size() > 0 ) {
		Note *note = midiNoteQueue[0];
		playingNotesQueue.push_back( note );		// aggiungo la nota alla lista di note da eseguire
		midiNoteQueue.pop_front();			// rimuovo la nota dalla lista di note

		// exclude notes
		Instrument *instr = note->getInstrument();
		if ( instr->excludeVect.size() != 0 ) {
			for (uint i = 0; i < instr->excludeVect.size(); i++) {
				Instrument *pExcluded = instr->excludeVect[ i ];
				// remove all notes listed in excludeVect
				for ( uint j = 0; j < playingNotesQueue.size(); j++ ) {	// delete older note
					Note *oldNote = playingNotesQueue[ j ];

					if ( oldNote->getInstrument() == pExcluded ) {
//						hydrogenInstance->infoLog("Delete excluded note");
						playingNotesQueue.erase( playingNotesQueue.begin() + j );
						delete oldNote;
					}
				}
			}
		}

		// raise noteOn event
		for ( uint nList = 0; nList < audioEngine_listeners.size(); nList++ ) {
			audioEngine_listeners[nList]->noteOn( note->copy() );	// the note must be deleted!
		}
	}
}





//----------------------------------------------------------------------------
/**
 * Update engine info
 */
inline void audioEngine_process_updateInfo() {
	playingNotes = playingNotesQueue.size();
}




/**
 *
 */
void audioEngine_seek( unsigned long nFrames )
{
	if (m_nTotalFrames == nFrames) {
		return;
	}

	char tmp[200];
	sprintf(tmp, "[audioEngine_seek()] seek in %d (old pos = %d)", nFrames, m_nTotalFrames);
	hydrogenInstance->infoLog(tmp);

	m_nTotalFrames = nFrames;

	int tickNumber_start = (uint)( m_nTotalFrames / tickSize );
//	sprintf(tmp, "[audioEngine_seek()] tickNumber_start = %d", tickNumber_start);
//	hydrogenInstance->infoLog(tmp);

	bool loop = m_pSong->isLoopEnabled();
	m_nPatternPos = findPatternInTick( tickNumber_start, loop, &m_nPatternStartTick );

//	sprintf(tmp, "[audioEngine_seek()] m_nPatternPos = %d", m_nPatternPos);
//	hydrogenInstance->infoLog(tmp);

	if ( m_pSong->getMode() == LIVE_MODE ) {
		m_pNextPattern = m_pSong->getPatternSequence()->get( m_nPatternPos );
	}

	audioEngine_clearNoteQueue();
}



/**
 *
 */
/**
 *
 */
inline void audioEngine_process_jackTransport()
{
#ifdef JACK_SUPPORT

	// skip if jack transport is disabled
	if ( (PreferencesMng::getInstance())->isJackTransportSlave() == false ) {
		return;
	}

	static double last_jack_bpm = 0;

	if ( m_pAudioDriver->getClassName() == "JackDriver" ) {
		if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING ) ) {
			m_pSong->setLoopEnabled( false );
		}

		JackDriver *jackDriver = ( (JackDriver*)m_pAudioDriver );
		jackDriver->updateTransportInfo();

		jack_transport_state_t state= jackDriver->getTransportState();
		jack_position_t position = jackDriver->getTransportPos();

		char tmp[300];

		switch (state) {
			case JackTransportStopped:
				// check current state
				if (audioEngine_state == PLAYING) {
					hydrogenInstance->infoLog( "----------" );
					hydrogenInstance->infoLog( "JackTransportStopped" );

					audioEngine_stop(false);	// no engine lock
					audioEngine_seek( position.frame );
					hydrogenInstance->infoLog( "----------" );
				}
				break;

			case JackTransportRolling:
				if (audioEngine_state == READY) {	// check current state
					if ( m_nTotalFrames != position.frame ) {
						hydrogenInstance->infoLog( "----------" );

						sprintf(tmp, "[audioEngine_process_jackTransport()] Rolling - position.frame = %d", position.frame);
						hydrogenInstance->infoLog(tmp);

//						audioEngine_start(false);	// no engine lock

						if (position.valid & JackPositionBBT) {
							last_jack_bpm = position.beats_per_minute;

							double bpm = position.beats_per_minute;
							double resolution = m_pSong->getResolution();
							double sampleRate = (float)m_pAudioDriver->getSampleRate();
							tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );

							m_pSong->setBpm( bpm );

							sprintf(tmp, "[audioEngine_process_jackTransport()] Rolling - ticksize = %d, bpm = %f", tickSize, bpm);
							hydrogenInstance->infoLog(tmp);
						}
						else{
							hydrogenInstance->warningLog( "JackPosition not valid" );
						}
						audioEngine_start( false, position.frame );	// no engine lock
						audioEngine_seek( position.frame );

/*
						/// test preso da audioEngine_start
						audioEngine_state = PLAYING;
						hydrogenInstance->infoLog("PLAYING");
						// raise stateChanged event
						for (uint i = 0; i < audioEngine_listeners.size(); i++) {
							audioEngine_listeners[i]->stateChanged(PLAYING);
						}
						// fine test
*/

						sprintf(tmp, "[audioEngine_process_jackTransport()] Rolling, m_nTotalFrames = %d", m_nTotalFrames);
						hydrogenInstance->infoLog(tmp);

						hydrogenInstance->infoLog( "----------" );
					}
				}
				else if ( audioEngine_state == PLAYING ) {

					if ( m_nTotalFrames != position.frame ) {
						hydrogenInstance->infoLog( "----------" );

						audioEngine_seek( position.frame );
						sprintf(tmp, "[audioEngine_process_jackTransport()] Rolling - m_nTotalFrames = %d", m_nTotalFrames);
						hydrogenInstance->infoLog(tmp);

						hydrogenInstance->infoLog( "----------" );
					}
					if (position.valid & JackPositionBBT) {
						if ( ((double)last_jack_bpm != (double)position.beats_per_minute) && (position.beats_per_minute != 0) ) {
							hydrogenInstance->infoLog( "----------" );

							hydrogenInstance->warningLog("!!!!![audioEngine_process_jackTransport()] Rolling");

//							audioEngine_stop(false);	// no engine lock
//							audioEngine_start(false);	// no engine lock
							cout << "[audioEngine_process_jackTransport()] Rolling - old ticksize = " << tickSize << endl;
							cout << "[audioEngine_process_jackTransport()] Rolling - new bpm = " << position.beats_per_minute << endl;

							double bpm = position.beats_per_minute;
							float resolution = m_pSong->getResolution();
							float sampleRate = (float)m_pAudioDriver->getSampleRate();
							tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );

							m_pSong->setBpm( bpm );

							last_jack_bpm = bpm;
							cout << "[audioEngine_process_jackTransport()] Rolling - new jack ticksize = " << tickSize << endl;

							hydrogenInstance->infoLog( "----------" );
						}
					}
				}
				break;

			case JackTransportStarting:
				if ( m_nTotalFrames != position.frame ) {
					hydrogenInstance->infoLog( "----------" );

					audioEngine_seek( position.frame );
					sprintf(tmp, "[audioEngine_process_jackTransport()] Starting - m_nTotalFrames = %d", m_nTotalFrames);
					hydrogenInstance->infoLog(tmp);

					hydrogenInstance->infoLog( "----------" );
				}
				if (position.valid & JackPositionBBT) {
					if ( ((double)last_jack_bpm != (double)position.beats_per_minute) && (position.beats_per_minute != 0) ) {
						hydrogenInstance->infoLog( "----------" );

						hydrogenInstance->infoLog("[audioEngine_process_jackTransport()] Starting");

//						audioEngine_stop(false);	// no engine lock
//						audioEngine_start(false);	// no engine lock
		//				cout << "[audioEngine_process_jackTransport()] Rolling - old ticksize = " << tickSize << endl;
		//				cout << "[audioEngine_process_jackTransport()] Rolling - new bpm = " << position.beats_per_minute << endl;

						double bpm = position.beats_per_minute;
						float resolution = m_pSong->getResolution();
						float sampleRate = (float)m_pAudioDriver->getSampleRate();
						tickSize = (uint)( sampleRate * 60.0 /  bpm / resolution );

						m_pSong->setBpm( bpm );

						last_jack_bpm = bpm;
						sprintf(tmp, "[audioEngine_process_jackTransport()] Starting - ticksize = %d, bpm = %f", tickSize, bpm);
						hydrogenInstance->infoLog(tmp);

						hydrogenInstance->infoLog( "----------" );
					}
				}
				break;

			default:
				hydrogenInstance->errorLog( "[audioEngine_process_jackTransport()] Unknown jack transport state" );
		}
	}

#endif // #ifdef JACK_SUPPORT

}




void audioEngine_clearNoteQueue() {
	hydrogenInstance->infoLog("[audioEngine_clearNoteQueue()]");

	// delete all copied notes in the song notes queue
	for (uint i = 0; i < songNoteQueue.size(); i++) {
		Note *note = songNoteQueue[i];
		delete note;
	}
	songNoteQueue.clear();

	// delete all copied notes in the playing notes queue
	for (uint i = 0; i < playingNotesQueue.size(); i++) {
		Note *note = playingNotesQueue[i];
		delete note;
	}
	playingNotesQueue.clear();

	// delete all copied notes in the midi notes queue
	for (uint i = 0; i < midiNoteQueue.size(); i++) {
		Note *note = midiNoteQueue[i];
		delete note;
	}
	midiNoteQueue.clear();

}





//----------------------------------------------------------------------------
/**
 *
 */
int audioEngine_process(uint32_t nframes, void *arg)
{
//	audioEngine_lock();

	// try to lock the engine mutex
	// if fails returns without sound processing
	int res = pthread_mutex_trylock(&m_engineLock_mutex);
	if (res != 0) {
		// update total frames number
//		m_nTotalFrames += nframes;	// FIXME: questo non mi convince...
		hydrogenInstance->warningLog( "audioEngine_process: trylock != 0" );
		return 0;
	}

	timeval startTimeval = currentTime2();

	audioEngine_process_jackTransport();

	// (re)create audio buffer
	if ( audioEngine_buffer_size != nframes ) {
		char tmp[200];
		sprintf(tmp, "Buffer size changed. old size = %d, new size = %d", audioEngine_buffer_size, nframes);
		hydrogenInstance->infoLog(tmp);

		audioEngine_buffer_size = nframes;

		if ( delayBuffer_L != NULL ) {
			delete[] delayBuffer_L;
		}
		if ( delayBuffer_R != NULL ) {
			delete[] delayBuffer_R;
		}
		if ( m_pFxBuffer_L ) {
			delete[] m_pFxBuffer_L;
		}
		if ( m_pFxBuffer_R ) {
			delete[] m_pFxBuffer_R;
		}

		delayBuffer_L = new float[ nframes ];
		delayBuffer_R = new float[ nframes ];

		if ( m_pLadspaFX ) {
			m_pFxBuffer_L = new float[ nframes ];
			m_pFxBuffer_R = new float[ nframes ];

			// update ladspa fx connections
			m_pLadspaFX->connectPorts( m_pFxBuffer_L, m_pFxBuffer_R, m_pFxBuffer_L, m_pFxBuffer_R );
		}
	}

	memset( main_buffer_L, 0, nframes * sizeof( float ) );	// clear main out Left
	memset( main_buffer_R, 0, nframes * sizeof( float ) );	// clear main out Right

	if (m_pLadspaFX) {
		memset( m_pFxBuffer_L, 0, nframes * sizeof( float ) );
		memset( m_pFxBuffer_R, 0, nframes * sizeof( float ) );
	}

	if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING) ) {
		if ( m_pSong->isDelayEnabled() ) {
			memset( delayBuffer_L, 0, nframes * sizeof( float ) );		// clear delay buffer Left
			memset( delayBuffer_R, 0, nframes * sizeof( float ) );		// clear delay buffer Right
		}
	}

	if (use_track_outs) {
		// clear track buffers
	 	for (unsigned int n=0; n < MAX_TRACKS; ++n) {
 			memset( track_buffers_L[n], 0, nframes * sizeof( float ) );
 			memset( track_buffers_R[n], 0, nframes * sizeof( float ) );
	 	}
	}

	if ( ( audioEngine_state == READY) || (audioEngine_state == PLAYING ) ) {
		audioEngine_process_checkBPMChanged();
	}

	// check ticksize
	if ( tickSize == 0 ) {
		hydrogenInstance->infoLog( "audioEngine_process: tickSize == 0" );
		if ( audioEngine_state != PREPARED ) {
			hydrogenInstance->errorLog( "tickSize == 0 audioEngine_state != PREPARED!" );
		}
	}
	audioEngine_process_updateInfo();

	bool sendPatternChange = false;
	if ( audioEngine_state == PLAYING ) {
		// update the notes queue
		int res = audioEngine_updateNoteQueue( nframes );
		if ( res == -1 ) {	// end of song
			hydrogenInstance->infoLog( "audioEngine_process: End of song received, calling engine_stop()" );
			audioEngine_unlock();
			audioEngine_stop();
			if ( m_pAudioDriver->getClassName() == "DiskWriterDriver" ) {
				hydrogenInstance->infoLog( "Exiting DiskWriterDriver thread!!" );
				return 1;	// kill the audio m_pAudioDriver thread
			}
		}
		else if ( res == 2 ) {	// send pattern change
			sendPatternChange = true;
		}
	}


	// play PATTERN
	audioEngine_process_playPattern( nframes );

	// MIDI EVENTS
	if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING) ) {
		audioEngine_process_playMidi(nframes);
	}

	// Max notes
	while ( playingNotesQueue.size() > maxNotes ) {
		Note *oldNote = playingNotesQueue[ 0 ];
		playingNotesQueue.erase( playingNotesQueue.begin() );
		delete oldNote;	// FIXME: send note-off instead of removing the note from the list?
	}


	// eseguo tutte le note nella lista di note in esecuzione
	uint i = 0;
	Note* note = NULL;
	while ( i < playingNotesQueue.size() ) {
		note = playingNotesQueue[ i ];		// recupero una nuova nota
		uint res = audioEngine_renderNote( note, nframes, nframes );
		if ( res == 1 ) {	// la nota e' finita
			playingNotesQueue.erase( playingNotesQueue.begin() + i );
			if ( note != NULL ) {
				delete note;
				note = NULL;
			}
		}
		else {
			i++; // carico la prox nota
		}
	}



	// Delay FX
	if ( ( audioEngine_state == READY ) || ( audioEngine_state == PLAYING) ) {
		float delay_wetLevel = m_pSong->getDelayWetLevel();
		uint delay_delayTime = tickSize * m_pSong->getDelayTime() * 3;
		float delay_feedbackLevel = m_pSong->getDelayFeedbackLevel();

		for ( uint i = 0; i < nframes; i++) {
			float val_L = main_buffer_L[i];
			float val_R = main_buffer_R[i];

			// Delay FX
			if ( m_pSong->isDelayEnabled() ) {
#ifdef DEBUG_MESSAGES
				if (delay_delayTime > MAX_DELAY) {
					hydrogenInstance->errorLog( "audioEngine_process: delayTime > MAX_DELAY" );
					delay_delayTime = MAX_DELAY;
				}
#endif
				int currentDelayPos = (delayStartPos + i) % delay_delayTime;

				float delayVal_L = delayRingBuffer_L[ currentDelayPos ]  * delay_wetLevel;
				float delayVal_R = delayRingBuffer_R[ currentDelayPos ] * delay_wetLevel;

				val_L += delayVal_L;
				val_R += delayVal_R;

				// feedback

				float temp_L = delayBuffer_L[i] + ( delayRingBuffer_L[ currentDelayPos ] * delay_feedbackLevel );
				float temp_R = delayBuffer_R[i] + ( delayRingBuffer_R[ currentDelayPos ] * delay_feedbackLevel );

				bool delay_pingPong = true;
				// ping pong delay
				if ( delay_pingPong == true ) {
					delayRingBuffer_R[ currentDelayPos ] = temp_L;
					delayRingBuffer_L[ currentDelayPos ] = temp_R;
				}
				else {
					delayRingBuffer_L[ currentDelayPos ] = temp_L;
					delayRingBuffer_R[ currentDelayPos ] = temp_R;
				}

				// update delayBuffer peak
				if ( delayVal_L > delayBufferPeak_L ) {
					delayBufferPeak_L = delayVal_L;
				}
				if ( delayVal_R > delayBufferPeak_R ) {
					delayBufferPeak_R = delayVal_R;
				}
			}

			main_buffer_L[i] = val_L;
			main_buffer_R[i] = val_R;
		}
		delayStartPos = (delayStartPos + nframes) % delay_delayTime;
	}



	// LADSPA test
	if ( m_pLadspaFX ) {
		m_pLadspaFX->processFX( nframes );
		for ( uint i = 0; i < nframes; i++) {
			main_buffer_L[ i ] += m_pFxBuffer_L[ i ];
			main_buffer_R[ i ] += m_pFxBuffer_R[ i ];
		}
	}



	for ( uint i = 0; i < nframes; i++) {
		float val_L = main_buffer_L[i];
		float val_R = main_buffer_R[i];
		// update master peak
		if (val_L > masterPeak_L) {
			masterPeak_L = val_L;
		}
		if (val_R > masterPeak_R) {
			masterPeak_R = val_R;
		}
	}

	// update total frames number
	if ( audioEngine_state == PLAYING ) {
		m_nTotalFrames += nframes;
	}

	timeval finishTimeval = currentTime2();
	processTime = (finishTimeval.tv_sec - startTimeval.tv_sec) * 1000.0 + (finishTimeval.tv_usec - startTimeval.tv_usec) / 1000.0;

	float sampleRate = (float)m_pAudioDriver->getSampleRate();
	maxProcessTime = 1000.0 / (sampleRate / nframes);

	//DEBUG
	if ( processTime > maxProcessTime ) {
		char xrunMsg[200];
		sprintf( xrunMsg, "XRUN of %f msec ( %f > %f ) - notes = %d", (processTime - maxProcessTime), processTime, maxProcessTime, playingNotes);
		hydrogenInstance->warningLog( string( xrunMsg ) );
		// raise xRun event
		for (uint i = 0; i < audioEngine_listeners.size(); ++i) {
			audioEngine_listeners[i]->xRunEvent( processTime - maxProcessTime, maxProcessTime );
		}
	}

	audioEngine_unlock();

	if ( sendPatternChange ) {
		// raise patternChanged event
		for (uint i = 0; i < audioEngine_listeners.size(); ++i) {
			audioEngine_listeners[i]->patternChanged();
		}
	}

	return 0;
}





//----------------------------------------------------------------------------
/** Funzione per il rendering di una singola nota
 *  Restituisce 0 se la nota non e' finita
 *  Restituisce 1 se la nota e' terminata
 */
inline uint audioEngine_renderNote(Note* note, int nFrames, uint bufferSize) {
	// nFrames = frame disponibili da copiare nei buffer

/*
	if (note->samplePosition == 0) {
		hydrogenInstance->infoLog( "note on" );
	}
*/
	Instrument *instr = note->getInstrument();
	if (instr == NULL) {
		hydrogenInstance->errorLog( "audioEngine_renderNote: Instrument Null" );
		return 0;
	}

	int track_n = m_pSong->getInstrumentList()->getPos(instr);

	Sample *s = instr->getSample();
	if (s == NULL) {
		hydrogenInstance->errorLog("audioEngine_renderNote: sample Null");
		return 0;
	}

	float *sample_data_L = s->getData_L();
	float *sample_data_R = s->getData_R();
	float instrPeak_L = instr->getPeak_L(); // this value will be reset to 0 by the mixer..
	float instrPeak_R = instr->getPeak_R(); // this value will be reset to 0 by the mixer..

	int noteStartInFrames = ( note->getPosition() * tickSize ) + note->getHumanizeDelay();

	int initialSilence = 0;
	if (noteStartInFrames > (int)m_nTotalFrames) {	// scrivo silenzio prima dell'inizio della nota
		initialSilence = noteStartInFrames - m_nTotalFrames;
		nFrames = nFrames - initialSilence;
		if (nFrames < 0) {	// delay note execution
			return 0;
		}
	}

	float cost_L = 1;
	float cost_R = 1;
	float cost_track = 1;

	if (instr->isMuted()) {	// is instrument muted?
		cost_L = 0.0;
		cost_R = 0.0;
	}
	else {	// Precalcolo i valori per le nota
		cost_L = cost_L * note->getVelocity();		// note velocity
		cost_L = cost_L * note->getPan_L();		// note pan
		cost_L = cost_L * instr->getPan_L();		// instrument pan
		cost_L = cost_L * instr->getVolume();		// instrument volume
		cost_L = cost_L * m_pSong->getVolume();		// song volume

		cost_R = cost_R * note->getVelocity();	// note velocity
		cost_R = cost_R * note->getPan_R();		// note pan
		cost_R = cost_R * instr->getPan_R();		// instrument pan
		cost_R = cost_R * instr->getVolume();		// instrument volume
		cost_R = cost_R * m_pSong->getVolume();	// song pan
	}

	// direct track outputs only use velocity
	cost_track = note->getVelocity();

	float delayLevel = instr->getDelayLevel();
//	float nFx1Level = 1.0;	// ladspa test

	// Se non devo fare resample (drumkit) posso evitare di utilizzare i float e gestire il tutto in
	// maniera ottimizzata

//	constant^12 = 2, so constant = 2^(1/12) =
   //     1.059463.

//	float nStep = 1.0;1.0594630943593

	float nStep = 1.0;
	if ((int)note->getPitch() == 0) {	// NO RESAMPLE
		int avail_bytes = s->getNFrames() - (int)note->samplePosition;	// verifico il numero di frame disponibili ancora da eseguire

		int retValue = 1; // the note is ended
		if (avail_bytes > (int)(bufferSize - initialSilence)) {	// il sample e' piu' grande del buffersize
			// imposto il numero dei bytes disponibili uguale al buffersize
			avail_bytes = bufferSize - initialSilence;
			retValue = 0; // the note is not ended yet
		}

		// copio il sample nel buffer
		float val_L;
		float val_R;
		int bufferPos = initialSilence;
		int samplePos = (int)note->samplePosition;
		int i;
		for (i = 0; i < avail_bytes; ++i) {
			val_L = sample_data_L[samplePos] * cost_L;
			val_R = sample_data_R[samplePos] * cost_R;

			main_buffer_L[bufferPos] += val_L;
			main_buffer_R[bufferPos] += val_R;

			delayBuffer_L[bufferPos] += val_L * delayLevel;
			delayBuffer_R[bufferPos] += val_R * delayLevel;

/*
			m_pFxBuffer_L[ bufferPos ] += val_L * nFx1Level;
			m_pFxBuffer_R[ bufferPos ] += val_R * nFx1Level;
*/
			if (use_track_outs && track_n >= 0) {	// for the track output
				track_buffers_L[track_n][bufferPos] = sample_data_L[samplePos] * cost_track;
				track_buffers_R[track_n][bufferPos] = sample_data_R[samplePos] * cost_track;
			}

			// update instr peak
			if (val_L > instrPeak_L) {	instrPeak_L = val_L;	}
			if (val_R > instrPeak_R) {	instrPeak_R = val_R;	}

			++samplePos; //samplePos += 1;
			++bufferPos;
		}
	//	note->samplePosition += i * nStep;
		note->samplePosition += i;

		instr->setPeak_L( instrPeak_L );
		instr->setPeak_R( instrPeak_R );

		return retValue;
	}
	else {	// RESAMPLE
		nStep = pow( 1.0594630943593, (int)note->getPitch() );
		int avail_bytes = (int)( (float)(s->getNFrames() - note->samplePosition) / nStep );	// verifico il numero di frame disponibili ancora da eseguire

		int retValue = 1; // the note is ended
		if (avail_bytes > (int)(bufferSize - initialSilence)) {	// il sample e' piu' grande del buffersize
			// imposto il numero dei bytes disponibili uguale al buffersize
			avail_bytes = bufferSize - initialSilence;
			retValue = 0; // the note is not ended yet
		}

		// copio il sample nel buffer
		float val_L;
		float val_R;
		float sample_L, sample_R;
		int bufferPos = initialSilence;
		float samplePos = note->samplePosition;
		int i;
		for (i = 0; i < avail_bytes; ++i) {
//8-11-2003	samplePos += nStep;

			// linear interpolation
			// new = old(int)+(old(int+1)-old(int))*fract
			// FIXME: error! samplepos+1 can exceed the array bounds
//			#warning "FIXME: error! samplepos+1 can exceed the array bounds"
			sample_L = sample_data_L[(int)samplePos] + ( sample_data_L[(int)(samplePos + 1)] - sample_data_L[(int)samplePos] ) * (samplePos - (int)samplePos);
			sample_R = sample_data_R[(int)samplePos] + ( sample_data_R[(int)(samplePos + 1)] - sample_data_R[(int)samplePos] ) * (samplePos - (int)samplePos);

			val_L = sample_L * cost_L;
			val_R = sample_R * cost_R;

			main_buffer_L[bufferPos] += val_L;
			main_buffer_R[bufferPos] += val_R;

			delayBuffer_L[bufferPos] += val_L * delayLevel;
			delayBuffer_R[bufferPos] += val_R * delayLevel;

			if (use_track_outs && track_n >= 0) {	// for the track output
				track_buffers_L[track_n][bufferPos] = sample_L * cost_track;
				track_buffers_R[track_n][bufferPos] = sample_R * cost_track;
			}

			// update instr peak
			if (val_L > instrPeak_L) {	instrPeak_L = val_L;	}
			if (val_R > instrPeak_R) {	instrPeak_R = val_R;	}

			samplePos += nStep;
			++bufferPos;
		}
		note->samplePosition += i * nStep;

		instr->setPeak_L( instrPeak_L );
		instr->setPeak_R( instrPeak_R );

		return retValue;
	}
}





//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_setSong(Song *newSong) {
	hydrogenInstance->infoLog( "set song \"" + newSong->getName() + "\"" );

	audioEngine_lock();

	// check current state
	if (audioEngine_state != PREPARED) {
		hydrogenInstance->errorLog("[audioEngine_setSong]: Error the audio engine is not in PREPARED state");
	}

	m_pPattern = NULL;
	m_pNextPattern = NULL;

	audioEngine_clearNoteQueue();

	m_pSong = newSong;

	// update ticksize
	audioEngine_process_checkBPMChanged();

	// find the first pattern and set as current
	if ( m_pSong->getPatternList()->getSize() > 0 ) {
		m_pPattern = m_pSong->getPatternList()->get(0);
	}

	// change the current audio engine state
	audioEngine_state = READY;
	hydrogenInstance->infoLog("READY");

	// raise patternChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->patternChanged();
	}

	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(READY);
	}

	audioEngine_unlock();
}



//----------------------------------------------------------------------------
/**
 *
 */
void audioEngine_removeSong() {
	audioEngine_lock();
	// check current state
	if (audioEngine_state != READY) {
		hydrogenInstance->errorLog("audioEngine_setSong: Error the audio engine is not in READY state");
		audioEngine_unlock();
		return;
	}

	m_pSong = NULL;
	m_pPattern = NULL;
	m_pNextPattern = NULL;


	audioEngine_clearNoteQueue();

	// change the current audio engine state
	audioEngine_state = PREPARED;
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(PREPARED);
	}

	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 *
 */
// return -1 = end of song
// return 2 = send pattern changed event!!
inline int audioEngine_updateNoteQueue(uint nFrames)
{
	static int lastTick = -1;
	bool sendPatternChange = false;

	int tickNumber_start = (uint)( m_nTotalFrames / tickSize );
	int tickNumber_end = (uint)( ( m_nTotalFrames + nFrames ) / tickSize );

	int tick = tickNumber_start;
	while ( tick <= tickNumber_end ) {
		if (tick == lastTick) {
			tick++;
			continue;
		}
		else {
			lastTick = tick;
		}

		// SONG MODE
		if ( m_pSong->getMode() == SONG_MODE ) {

			m_nPatternPos = findPatternInTick( tick, false, &m_nPatternStartTick );
			m_nPatternTickPosition = tick - m_nPatternStartTick;

			if (m_nPatternTickPosition == 0) {
				sendPatternChange = true;
			}

			if ( m_nPatternPos == -1 ) {
//				hydrogenInstance->infoLog("[audioEngine_updateNoteQueue()] pattern pos = -1");

				if ( m_pSong->isLoopEnabled() == true ) {	// check if the song loop is enabled
					if ( m_pSong->getPatternSequence()->getSize() != 0 ) {
						if ( m_nPatternPos == -1 ) {
							// Loop mode!!
							hydrogenInstance->infoLog("[audioEngine_updateNoteQueue()] Loop");
							int nLoopStartTick = 0;
							m_nPatternPos = findPatternInTick( nLoopStartTick, false, &m_nPatternStartTick );
							audioEngine_seek(m_nPatternStartTick * tickSize);
							return audioEngine_updateNoteQueue(m_nTotalFrames);	// recursive...
						}
					}
				}
				else {
					hydrogenInstance->infoLog( "[audioEngine_updateNoteQueue()] End of Song" );
					return -1;
				}
			}
			m_pPattern = m_pSong->getPatternSequence()->get( m_nPatternPos );



			if (m_nPatternTickPosition > MAX_NOTES) {
				char tmp[200];
				sprintf(tmp, "[audioEngine_updateNoteQueue()]1 Error m_nPatternTickPosition > MAX_NOTES. (%d > %d)", m_nPatternTickPosition, MAX_NOTES);
				hydrogenInstance->errorLog(tmp);

				sprintf(tmp, "[audioEngine_updateNoteQueue()] tick =%d, m_nPatternStartTick = %d, m_nPatternPos = %d", tick, m_nPatternStartTick, m_nPatternPos);
				hydrogenInstance->errorLog(tmp);

				// PANIC!!!
				hydrogenInstance->errorLog("[audioEngine_updateNoteQueue()] Panic! Stopping audio engine");
				audioEngine_stop(false);	// stop the audio engine
				return 0;
			}


		}

		// LIVE MODE
		else if ( m_pSong->getMode() == LIVE_MODE )	{
			if ( ( tick == (int)(m_nPatternStartTick + m_pPattern->getSize()) ) || ( m_nPatternStartTick == -1 ) ) {
				if ( m_pNextPattern != NULL ) {
					m_pPattern = m_pNextPattern;
					m_pNextPattern = NULL;
					sendPatternChange = true;
				}
				m_nPatternStartTick = tick;
			}
			m_nPatternTickPosition = tick - m_nPatternStartTick;
		}

		// metronome
		if ( ( tick % 48 ) == 0) {
			if (metronomeInstrument->getVolume() != 0.0) {
				Note *metronomeNote = new Note( tick, 1.0, 1.0, 1.0 );
				metronomeNote->setInstrument( metronomeInstrument );
				audioEngine_noteOn( metronomeNote );
			}
		}

		// update notes queue
		if ( m_pPattern != NULL ) {
			uint nSequences = ( m_pPattern->getSequenceList() )->getSize();
			//uint nNote = tick % MAX_NOTES;
			uint nNote = m_nPatternTickPosition;

			if (m_nPatternTickPosition > MAX_NOTES) {
				char tmp[200];
				sprintf(tmp, "[audioEngine_updateNoteQueue()] Error m_nPatternTickPosition > MAX_NOTES. (%d > %d)", m_nPatternTickPosition, MAX_NOTES);
				hydrogenInstance->errorLog(tmp);

				// PANIC!!!
				hydrogenInstance->errorLog("[audioEngine_updateNoteQueue()] Panic! Stopping audio engine");
				audioEngine_stop(false);	// stop the audio engine
				return 0;
			}

			for ( uint i = 0; i < nSequences; i++ ) {
				Sequence* sequence = ( (m_pPattern->getSequenceList() )->get(i) );
				Note* note = sequence->noteList[ nNote ];
				if ( note ) {
					uint offset = 0;
					float velocity = note->getVelocity();

					// Swing
					if ( m_pSong->isSwingEnabled() ) {
						float swingFactor = m_pSong->getSwingFactor();
						if ( ( (nNote % 12) == 0 ) && ( (nNote % 24) != 0) ) {	// da l'accento al tick 4, 12, 20, 36...
							offset += (int)( ( 2.0 * tickSize ) * swingFactor );
						}
					}

					// Humanize - Time parameter
					if ( m_pSong->isHumanizeTimeEnabled() ) {
						int maxTimeHumanize = 2000;
						if (m_pSong->getHumanizeTimeValue() != 0) {
							float random = randomValue(100) / 100.0;
							offset += (int)(random * m_pSong->getHumanizeTimeValue() * maxTimeHumanize);
						}
					}

					// Humanize - Velocity parameter
					if ( m_pSong->isHumanizeVelocityEnabled() ) {
						if (m_pSong->getHumanizeVelocityValue() != 0) {
							float random = m_pSong->getHumanizeVelocityValue() * (randomValue(100) / 100.0);
							velocity += (random - (m_pSong->getHumanizeVelocityValue() / 2.0));
							if (velocity > 1.0) {
								velocity = 1.0;
							}
							else if (velocity < 0.0) {
								velocity = 0.0;
							}
						}
					}

					Note *copiedNote = new Note( note->getPosition(), velocity, note->getPan_L(), note->getPan_R(), note->getPitch() );
					copiedNote->setInstrument( note->getInstrument() );
					copiedNote->setPosition( tick );
					copiedNote->samplePosition = 0;
					copiedNote->setHumanizeDelay( offset );	// humanize time
					songNoteQueue.push_back( copiedNote );
				}
			}
		}
		tick++;
	}


	// audioEngine_process must send the pattern change event after mutex unlock
	if (sendPatternChange) {
		return 2;
	}
	return 0;
}




/**
 * Find the pattern for tick number
 */
inline int findPatternInTick( int tick, bool loopMode,int *pPatternStartTick  )
{
	if (m_pSong == NULL) {
		return -1;
	}

//	cout << "find pattern in tick " << tick << " [lastTick=" << findPatternInTick_lastTick << "] lastTotal[" << findPatternInTick_lastTotal << "] "	 << endl;
	int pos = -1;

	int totalTick = 0;
	int nPatterns = m_pSong->getPatternSequence()->getSize();
	int patternSize = 0;



/*
	// vecchia ricerca
	for ( int i = 0; i < nPatterns; i++ ) {
		cout << i << endl;
		patternSize = m_pSong->getPatternSequence()->get( i )->getSize();
		if ( (tick >= totalTick) && ( tick < totalTick + patternSize ) ) {
			findPatternInTick_lastTick = i;
			return i;
		}
		totalTick += patternSize;
	}
	cout << "ho finito con la vecchia ricerca" << endl;
*/


	totalTick = findPatternInTick_lastTotal;
	// nuova ricerca semi-ottimizzata
	Pattern *pattern = NULL;
	for ( int i = findPatternInTick_lastTick; i < nPatterns; i++ ) {
//		cout << i << endl;
		pattern = m_pSong->getPatternSequence()->get( i );
		if (pattern == NULL) {
			hydrogenInstance->errorLog( "[Hydrogen::findPatternInTick] Pattern == null");
			return -1;
		}
		patternSize = pattern->getSize();

		if ( (tick >= totalTick) && ( tick < totalTick + patternSize ) ) {
			findPatternInTick_lastTick = i;
			findPatternInTick_lastTotal = totalTick;
			(*pPatternStartTick) = totalTick;
			return i;
		}
		totalTick += patternSize;
	}


	if ( loopMode ) {
//		hydrogenInstance->infoLog( "[Hydrogen::findPatternInTick] loop mode" );
		// se arriva qua e' attivo il loop mode
		int songTick = totalTick;

		int loopTick = 0;
		if (songTick != 0) {
			loopTick = tick % songTick;
		}

		totalTick = 0;
		for ( int i = 0; i < nPatterns; i++ ) {
			patternSize = m_pSong->getPatternSequence()->get( i )->getSize();
			if ( ( loopTick >= totalTick ) && ( loopTick < totalTick + patternSize ) ) {
				findPatternInTick_lastTick = i;
				findPatternInTick_lastTotal = totalTick;
				(*pPatternStartTick) = totalTick;
				return i;
			}
			totalTick += patternSize;
		}
	}

	findPatternInTick_lastTick = 0;
	totalTick = 0;
	(*pPatternStartTick) = totalTick;
	return pos;
}




//----------------------------------------------------------------------------
/**
 *
 */
void	audioEngine_noteOn(Note *note) {
	// check current state
	if ( (audioEngine_state != READY) && (audioEngine_state != PLAYING) ) {
		hydrogenInstance->errorLog("audioEngine_noteOn: Error the audio engine is not in READY state");
		delete note;
		return;
	}

	midiNoteQueue.push_back(note);
}




//----------------------------------------------------------------------------
/**
 *
 */
void	audioEngine_noteOff( Note *note )
{
	if (note == NULL)	{
		hydrogenInstance->errorLog("audioEngine_noteOff: Error, note == NULL");
	}

	audioEngine_lock();

	// check current state
	if ( (audioEngine_state != READY) && (audioEngine_state != PLAYING) ) {
		hydrogenInstance->errorLog("audioEngine_noteOff: Error the audio engine is not in READY state");
		delete note;
		audioEngine_unlock();
		return;
	}

	for ( uint i = 0; i < playingNotesQueue.size(); i++ ) {	// delete old note
		Note *oldNote = playingNotesQueue[ i ];

		if ( oldNote->getInstrument() == note->getInstrument() ) {
			playingNotesQueue.erase( playingNotesQueue.begin() + i );
			delete oldNote;
			break;
		}
	}
	audioEngine_unlock();

	delete note;
}




//----------------------------------------------------------------------------
/**
 *
 */
unsigned long audioEngine_getTickPosition() {
	return m_nPatternTickPosition;
}





//----------------------------------------------------------------------------
/**
 * Start all audio drivers
 */
void audioEngine_startAudioDrivers() {
	audioEngine_lock();

	hydrogenInstance->infoLog("Starting audio drivers");

	// check current state
	if (audioEngine_state != INITIALIZED) {
		hydrogenInstance->errorLog("audioEngine_startAudioDrivers: Error the audio engine is not in INITIALIZED state");
		audioEngine_unlock();
		return;
	}

	// check if the audio m_pAudioDriver is still alive
	if (m_pAudioDriver) {
		hydrogenInstance->errorLog( "The audio driver is still alive" );
	}


	PreferencesMng *preferencesMng = PreferencesMng::getInstance();

	use_track_outs = false;

	// Create Audio Driver
	string audioDriver = preferencesMng->getAudioDriver();
	if (audioDriver == "Oss") {
#ifdef OSS_SUPPORT
		m_pAudioDriver = new OssDriver( audioEngine_process );
#else
		audioEngine_raiseError( Hydrogen::NO_OSS_SUPPORT );
		m_pAudioDriver = new NullDriver( audioEngine_process );
#endif
	}
	else if (audioDriver == "Jack") {
#ifdef JACK_SUPPORT
		m_pAudioDriver = new JackDriver(audioEngine_process);

		// set use track outputs flag
		use_track_outs = preferencesMng->isJackTrackOuts();
		((JackDriver *) m_pAudioDriver)->setTrackOuts( use_track_outs );

		use_default_outs = preferencesMng->isJackConnectDefaults();
		((JackDriver *) m_pAudioDriver)->setConnectDefaults( use_default_outs );
#else
		audioEngine_raiseError( Hydrogen::NO_JACK_SUPPORT );
		m_pAudioDriver = new NullDriver( audioEngine_process );
#endif
	}
	else {
		audioEngine_raiseError( Hydrogen::UNKNOWN_DRIVER );
	}

	// initialize the audio driver
	int res = m_pAudioDriver->init( preferencesMng->getBufferSize() );
	if (res != 0) {
		audioEngine_raiseError( Hydrogen::ERROR_STARTING_DRIVER );
		hydrogenInstance->errorLog("Error starting audio driver [audioDriver::init()]");
		delete m_pAudioDriver;
		use_track_outs = false;

		m_pAudioDriver = new NullDriver( audioEngine_process );
		m_pAudioDriver->init(0);
	}

	main_buffer_L = m_pAudioDriver->getOut_L();
	main_buffer_R = m_pAudioDriver->getOut_R();

#ifdef JACK_SUPPORT
	if (use_track_outs) {
		for (uint i = 0; i < MAX_TRACKS; ++i) {
			track_buffers_L[i] = ((JackDriver*)m_pAudioDriver)->getTrackOut_L(i);
			track_buffers_R[i] = ((JackDriver*)m_pAudioDriver)->getTrackOut_R(i);
		}
	}
#endif

	// LADSPA TEST
//	m_pLadspaFX = new LadspaFX( "/usr/lib/ladspa/amp.so", "amp_stereo" );
//	m_pLadspaFX = new LadspaFX( "/usr/lib/ladspa/alias_1407.so", "alias" );
//	m_pLadspaFX = new LadspaFX( "/usr/lib/ladspa/cmt.so", "canyon_delay" );
//	m_pLadspaFX = new LadspaFX( "/usr/lib/ladspa/cmt.so", "freeverb3" );
//	m_pLadspaFX = new LadspaFX( "/usr/lib/ladspa/lcr_delay_1436.so", "lcrDelay" );

	if ( m_pLadspaFX ){
		// m_pFxBuffer_L here is NULL
//		m_pLadspaFX->connectPorts( m_pFxBuffer_L, m_pFxBuffer_R, main_buffer_L, main_buffer_R );
		m_pLadspaFX->activate();
	}



// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ

	// check if midi driver is still alive
	if (midiDriver) {
		hydrogenInstance->errorLog( "The MIDI driver is still alive");
	}

	// Create MIDI driver
	midiDriver = new AlsaMidiDriver();
	midiDriver->open();
	midiDriver->setActive(true);
#endif

	res = m_pAudioDriver->connect();
	if (res != 0) {
		audioEngine_raiseError( Hydrogen::ERROR_STARTING_DRIVER );
		hydrogenInstance->errorLog("Error starting audio driver [audioDriver::connect()]");

		delete m_pAudioDriver;
		use_track_outs = false;
		m_pAudioDriver = new NullDriver( audioEngine_process );
		m_pAudioDriver->init(0);
		m_pAudioDriver->connect();
	}

	// change the current audio engine state
	if (m_pSong == NULL) {
		audioEngine_state = PREPARED;
		hydrogenInstance->infoLog("audioEngine_startAudioDrivers: PREPARED");

	}
	else {
		audioEngine_state = READY;
		hydrogenInstance->infoLog("READY");
	}
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(audioEngine_state);
	}
	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 * Stop all audio drivers
 */
void audioEngine_stopAudioDrivers() {
	hydrogenInstance->infoLog("Stopping audio drivers");

	// check current state
	if ( (audioEngine_state != PREPARED) && (audioEngine_state != READY) ) {
		hydrogenInstance->errorLog("audioEngine_stopAudioDrivers: Error the audio engine is not in PREPARED or READY state");
		return;
	}

// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ
	hydrogenInstance->infoLog("Stopping MIDI driver");
	// delete MIDI driver
	if (midiDriver) {
		midiDriver->close();
		delete midiDriver;
		midiDriver = NULL;
	}
#endif

	// delete audio driver
	if (m_pAudioDriver) {
		m_pAudioDriver->disconnect();
		delete m_pAudioDriver;
		m_pAudioDriver = NULL;
	}


	if (m_pLadspaFX) {
		m_pLadspaFX->deactivate();
		delete m_pLadspaFX;
		m_pLadspaFX = NULL;
	}

	audioEngine_lock();
	// change the current audio engine state
	audioEngine_state = INITIALIZED;
	hydrogenInstance->infoLog("INITIALIZED");
	// raise stateChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->stateChanged(INITIALIZED);
	}
	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 * Restart all audio and midi drivers
 */
void audioEngine_restartAudioDrivers() {
	audioEngine_stopAudioDrivers();
	audioEngine_startAudioDrivers();
}






//----------------------------------------------------------------------------
//
// Implementation of Hydrogen class
//
//----------------------------------------------------------------------------






//----------------------------------------------------------------------------
/**
 * Constructor
 */
Hydrogen::Hydrogen(EngineListener *mainListener) {
	if (instance) {
		errorLog( "Hydrogen audio engine is already running" );
	}

	hydrogenInstance = this;
	instance = this;
	audioEngine_init();
	addEngineListener( mainListener );
	audioEngine_startAudioDrivers();
}





//----------------------------------------------------------------------------
/**
 * Destructor
 */
Hydrogen::~Hydrogen() {
	infoLog( "[~Hydrogen]" );
	if ( audioEngine_state == PLAYING ) {
		audioEngine_stop();
	}
	removeSong();
	audioEngine_stopAudioDrivers();
	audioEngine_destroy();
	instance = NULL;
}





//----------------------------------------------------------------------------
/**
 * Return the Hydrogen instance
 */
Hydrogen* Hydrogen::getInstance(EngineListener *listener) {
	if (instance == NULL) {
		instance = new Hydrogen( listener );
	}
	return instance;
}




//----------------------------------------------------------------------------
/**
 * Start the internal sequencer
 */
int Hydrogen::start() {
	return audioEngine_start();
}





//----------------------------------------------------------------------------
/**
 * Stop the internal sequencer
 */
void Hydrogen::stop() {
	audioEngine_stop();
}





//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::setSong(Song *newSong) {
	audioEngine_setSong(newSong);
}




//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::removeSong() {
	audioEngine_removeSong();
}


//----------------------------------------------------------------------------
/**
 *
 */
Song* Hydrogen::getSong() {
	return m_pSong;
}





//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::noteOn(Note *note) {
	audioEngine_noteOn(note);
}





//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::noteOff(Note *note) {
	audioEngine_noteOff(note);
}




//----------------------------------------------------------------------------
/**
 *
 */
float Hydrogen::getMasterPeak_L() {
	return masterPeak_L;
}




//----------------------------------------------------------------------------
/**
 *
 */
float Hydrogen::getMasterPeak_R() {
	return masterPeak_R;
}





//----------------------------------------------------------------------------
/**
 *
 */
unsigned long Hydrogen::getTickPosition() {
	return audioEngine_getTickPosition();
}






//----------------------------------------------------------------------------
/**
 *
 */
Pattern* Hydrogen::getCurrentPattern() {
	return m_pPattern;
}








//----------------------------------------------------------------------------
/**
 * Set the next pattern (Live mode only)
 */
void Hydrogen::setNextPattern( int pos )
{
	audioEngine_lock();

	if ( m_pSong->getMode() == LIVE_MODE ) {
		PatternList *patternList = m_pSong->getPatternList();
		if ( (pos >= 0) && ( pos < (int)patternList->getSize() ) ) {
			m_pNextPattern = patternList->get(pos);
		}
		else {
			errorLog("setNextPattern: pos not in patternList range");
			cout << "pos = " << pos << endl;
			cout << "patternListSize = " << patternList->getSize() << endl;
		}
	}
	else {
		errorLog("can't set next pattern in song mode");
	}

	audioEngine_unlock();
}




//----------------------------------------------------------------------------
/**
 *
 */
uint Hydrogen::getPatternPos()
{
	return m_nPatternPos;
}



//----------------------------------------------------------------------------
/**
 *
 */
uint Hydrogen::getPlayingNotes()
{
	return playingNotes;
}




//----------------------------------------------------------------------------
/**
 *
 */
void Hydrogen::restartDrivers()
{
	audioEngine_restartAudioDrivers();
}




//----------------------------------------------------------------------------
/**
 * Export a song to a wav file, returns the elapsed time in mSec
 */
uint Hydrogen::exportSong(string filename)
{
	if ( getState() == PLAYING ) {
		stop();
	}

	int oldEngineMode = m_pSong->getMode();
	m_pSong->setMode( SONG_MODE );

	bool oldLoopEnabled = m_pSong->isLoopEnabled();
	m_pSong->setLoopEnabled( false );

	PreferencesMng *preferencesMng = PreferencesMng::getInstance();

	// stop all audio drivers
	audioEngine_stopAudioDrivers();

	/*
		FIXME: Questo codice fa davvero schifo....
	*/
	// reset
	m_nTotalFrames = 0;	// reset total frames
	m_nPatternPos = -1;
	m_nPatternTickPosition = 0;
	audioEngine_state = PLAYING;
	m_nPatternStartTick = -1;
	findPatternInTick_lastTotal = 0;
	findPatternInTick_lastTick = 0;

	m_pAudioDriver = new DiskWriterDriver(audioEngine_process);

	( (DiskWriterDriver*)m_pAudioDriver )->setFilename(filename);

	int res = m_pAudioDriver->init(preferencesMng->getBufferSize());
	if (res != 0) {
		errorLog("Error starting disk writer driver [DiskWriterDriver::init()]");
	}

	main_buffer_L = m_pAudioDriver->getOut_L();
	main_buffer_R = m_pAudioDriver->getOut_R();

	uint start = currentTime();


	res = m_pAudioDriver->connect();
	if (res != 0) {
		errorLog("Error starting disk writer driver [DiskWriterDriver::connect()]");
	}

	m_pAudioDriver->disconnect();

	audioEngine_state = INITIALIZED;
	delete m_pAudioDriver;
	m_pAudioDriver = NULL;

	main_buffer_L = NULL;
	main_buffer_R = NULL;

	uint finish = currentTime();
	uint elapsedTime = finish - start;

	char tmpbuf[200];
	sprintf(tmpbuf, "Exported in %d msec", elapsedTime);
	hydrogenInstance->infoLog( tmpbuf );

	m_pSong->setMode( oldEngineMode );
	m_pSong->setLoopEnabled( oldLoopEnabled );

	m_nTotalFrames = 0;	// reset total frames
	m_nPatternPos = -1;
	m_nPatternTickPosition = 0;
	audioEngine_startAudioDrivers();

	return elapsedTime;
}





/**
 *
 */
uint Hydrogen::getSongNotesQueue() {
	return songNoteQueue.size();
}



/**
 * Used for display audio driver info
 */
GenericDriver* Hydrogen::getAudioDriver() {
	return m_pAudioDriver;
}




/**
 * Used for display midi driver info
 */
AlsaMidiDriver* Hydrogen::getMidiDriver() {
// check if ALSA_SEQ support is enabled
#ifdef USE_ALSA_SEQ
	return midiDriver;
#else
	return NULL;
#endif
}




/**
 *
 */
void Hydrogen::setMasterPeak_L(float value) {
	masterPeak_L = value;
}





/**
 *
 */
void Hydrogen::setMasterPeak_R(float value) {
	masterPeak_R = value;
}


/**
 *
 */
int Hydrogen::getState() {
	return audioEngine_state;
}




/**
 *
 */
void Hydrogen::setCurrentPattern(Pattern *newPattern) {
	audioEngine_lock();

	m_pPattern = newPattern;
	// raise patternChanged event
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->patternChanged();
	}

	audioEngine_unlock();
}




/**
 *
 */
void Hydrogen::addEngineListener(EngineListener *newListener) {
	audioEngine_lock();

	if (newListener) {
		audioEngine_listeners.push_back(newListener);
	}

	audioEngine_unlock();
}




/**
 * Lock the audio engine
 */
void Hydrogen::lockEngine() {
	audioEngine_lock();
}




/**
 * Unlock the audio engine
 */
void Hydrogen::unlockEngine() {
	audioEngine_unlock();
}




/**
 *
 */
float Hydrogen::getProcessTime() {
	return processTime;
}



/**
 *
 */
float Hydrogen::getMaxProcessTime() {
	return maxProcessTime;
}



/**
 *
 */
float Hydrogen::getDelayBufferPeak_L() {
	return delayBufferPeak_L;
}



/**
 *
 */
float Hydrogen::getDelayBufferPeak_R() {
	return delayBufferPeak_R;
}



/**
 *
 */
void Hydrogen::setDelayBufferPeak_L( float value ) {
	delayBufferPeak_L = value;
}



/**
 *
 */
void Hydrogen::setDelayBufferPeak_R( float value ) {
	delayBufferPeak_R = value;
}




/**
 *
 */
int Hydrogen::loadDrumkit( DrumkitInfo *drumkitInfo )
{
	infoLog( "loadDrumkit" );

	LocalFileMng fileMng;
	string drumkitPath = fileMng.getDrumkitDirectory( drumkitInfo->getName() );


	InstrumentList *songInstrList = m_pSong->getInstrumentList();
	InstrumentList *pKitInstrList = drumkitInfo->getInstrumentList();
	for (uint i = 0; i < pKitInstrList->getSize(); i++ ) {

		Instrument *newInstr = pKitInstrList->get( i );

		string filename = drumkitPath + drumkitInfo->getName() + "/" + newInstr->getSample()->getFilename();

		Instrument *oldInstr = songInstrList->get( i );

		Sample *newSample = Sample::load( filename );
		if (newSample == NULL) { // is sample valid?
			errorLog( "Error Loading drumkit: NULL sample, now using /emptySample.wav" );
			newSample = Sample::load( string(DATA_DIR) + "/emptySample.wav" );
			newSample->setFilename( string(DATA_DIR) + "/emptySample.wav" );

		}

		audioEngine_lock();

		// delete old sample
		Sample *oldSample = oldInstr->getSample();
		delete oldSample;

		// insert new sample from newInstrument
		newSample->setFilename( newInstr->getSample()->getFilename() );	// restore the "relative path" to drumkit
		oldInstr->setSample( newSample );

		// update instrument properties
		oldInstr->setName( newInstr->getName() );
		oldInstr->setPan_L( newInstr->getPan_L() );
		oldInstr->setPan_R( newInstr->getPan_R() );
		oldInstr->setVolume( newInstr->getVolume() );
		oldInstr->setDrumkitName( newInstr->getDrumkitName() );
		oldInstr->setMuted( newInstr->isMuted() );

		oldInstr->excludeVectId.clear();
		for (uint i = 0; i < newInstr->excludeVectId.size(); i++) {
			oldInstr->excludeVectId.push_back( newInstr->excludeVectId[ i ] );
		}
		oldInstr->excludeVect.clear();

		audioEngine_unlock();
	}
	audioEngine_lock();
	// rebuild the exclude vector
	for (uint nInstr = 0; nInstr < songInstrList->getSize(); nInstr++) {
		Instrument* pInstr = songInstrList->get( nInstr );
		for (uint i = 0; i < pInstr->excludeVectId.size(); i++) {
			int id = pInstr->excludeVectId[ i ];
			Instrument* pExcluded = songInstrList->get( id );
			pInstr->excludeVect.push_back( pExcluded );
		}
	}
	audioEngine_unlock();


	return 0;	//ok
}



/**
 *
 */
void Hydrogen::raiseError( uint nErrorCode )
{
	audioEngine_raiseError( nErrorCode );
}




/**
 * Used by AlsaMidiDriver
 */
void Hydrogen::raiseMidiActivityEvent()
{
	for (uint i = 0; i < audioEngine_listeners.size(); i++) {
		audioEngine_listeners[i]->midiActivityEvent();
	}
}



/**
 *
 */
unsigned long Hydrogen::getTotalFrames()
{
	return m_nTotalFrames;
}




/**
 * Set the position in the song
 */
void Hydrogen::setPatternPos( uint pos )
{
	audioEngine_lock();

	uint nPatterns = m_pSong->getPatternSequence()->getSize();
	if ( pos >= nPatterns ) {
		hydrogenInstance->warningLog( "[Hydrogen::setPatternPos()] patternPos > nPatterns" );
		audioEngine_unlock();
		return;
	}


	int totalTick = 0;
	int patternSize;
	Pattern *pattern = NULL;
	for ( uint i = 0; i < pos; i++ ) {
		pattern = m_pSong->getPatternSequence()->get( i );
		patternSize = pattern->getSize();

		totalTick += patternSize;
	}

	audioEngine_seek( totalTick * tickSize );

	audioEngine_unlock();
}



