/*
 * Hydrogen
 * Copyright(c) 2002-2003 by Alex >Comix< Cominu [comix@users.sourceforge.net]
 *
 * http://hydrogen.sourceforge.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: PreferencesMng.cpp,v 1.1 2003/12/09 11:21:44 comix Exp $
 *
 */

#include "PreferencesMng.h"

#include "tinyxml.h"
#include "config.h"


PreferencesMng* PreferencesMng::instance = NULL;



//----------------------------------------------------------------------------
/**
 * Return an instance of PreferencesMng
 */
PreferencesMng* PreferencesMng::getInstance() {
	if (instance == NULL) {
		instance = new PreferencesMng();
	}

	return instance;
}




//----------------------------------------------------------------------------
/**
 * Constructor
 */
PreferencesMng::PreferencesMng() {
	infoLog( "INIT" );

	// default values
	bufferSize = 1024;
	setAudioDriver("Oss");
	sampleRate = 44100;
	useMetronome = false;
	metronomeVolume = 0.5f;
	interfaceMode = "Child frame";

	applicationFontFamily = "Helvetica";
	applicationFontPointSize = 10;

	mixerFontFamily = "Helvetica";
	mixerFontPointSize = 8;
	m_nPatternEditorGridHeight = 15;

	demoPath = string(DATA_DIR) + "/demo_songs/";
	restoreLastSong = true;
	lastSongFilename = "";
	hearNewNotes = false;
	maxNotes = 128;

	mixerFalloffSpeed = FALLOFF_NORMAL;	// normal speed

	ossDevice = "/dev/dsp";

	jackPortName1 = "alsa_pcm:playback_1";
	jackPortName2 = "alsa_pcm:playback_2";
	jackTransportSlave = false;
	jackConnectDefaults = true;
	jackTrackOuts = true;

	midiPortChannel = -1;
	midiDest_name = "-";
	midiDest_client = -1;
	midiDest_port = -1;
	m_bIgnoreMidiNoteOff = true;

	m_nPatternEditorGridResolution = 8;
	m_bPatternEditorUsingTriplets = false;

	m_bShowInstrumentPeaks = true;

	// window properties
	mainFormProperties.x = 0;
	mainFormProperties.y = 0;
	mainFormProperties.width = 723;
	mainFormProperties.height = 645;

	mixerProperties.x = 10;
	mixerProperties.y = 350;
	mixerProperties.width = 829;
	mixerProperties.height = 220;

	patternEditorProperties.x = 280;
	patternEditorProperties.y = 100;
	patternEditorProperties.width = 706;
	patternEditorProperties.height = 302;

	songEditorProperties.x = 10;
	songEditorProperties.y = 10;
	songEditorProperties.width = 600;
	songEditorProperties.height = 250;

	drumkitManagerProperties.x = 500;
	drumkitManagerProperties.y = 20;

	audioEngineInfoProperties.x = 720;
	audioEngineInfoProperties.y = 120;


	loadPreferences();
}





//----------------------------------------------------------------------------
/**
 * Destructor
 */
PreferencesMng::~PreferencesMng() {
	savePreferences();
	infoLog( "DESTROY" );
	instance = NULL;
}




//----------------------------------------------------------------------------
/**
 * Get user's home directory.
 */
string PreferencesMng::getUserHome() {
	string home;
	struct passwd *pwdbuf;

	pwdbuf = getpwuid(getuid());
	home = pwdbuf->pw_dir;

	return home;
}




//----------------------------------------------------------------------------
/**
 * Load the preferences file
 */
void PreferencesMng::loadPreferences() {
	bool recreate = false;	// configuration file must be recreated?
	string prefDir = getUserHome() + "/.hydrogen";
	string dataDir = getUserHome() + "/.hydrogen/data";
	string filename = prefDir + "/hydrogen.conf";

	infoLog("Loading preferences file in " + filename);

	// pref directory exists?
	DIR* dirp = opendir( prefDir.c_str() );
	if ( dirp == NULL ) {
		warningLog( "Configuration directory not found." );
		createPreferencesDirectory();
	}

	// data directory exists?
	DIR* dirp2 = opendir( dataDir.c_str() );
	if ( dirp2 == NULL ) {
		warningLog( "Data directory not found." );
		createDataDirectory();
	}


	// pref file exists?
	std::ifstream input(filename.c_str() , std::ios::in | std::ios::binary);
	if (input){
		// read preferences file
		TiXmlDocument doc(filename.c_str());
		doc.LoadFile();

		TiXmlNode* rootNode;
		if ( (rootNode = doc.FirstChild("hydrogen_preferences")) ) {

			// version
			string version = readXmlString( rootNode, "version", "" );
			if ( version == "" ) {
				recreate = true;
			}

			//////// GENERAL ///////////
			restoreLastSong = readXmlBool( rootNode, "restoreLastSong", restoreLastSong );
			hearNewNotes = readXmlBool( rootNode, "hearNewNotes", hearNewNotes );

			/////////////// AUDIO ENGINE //////////////
			TiXmlNode* audioEngineNode;
			if ( !(audioEngineNode = rootNode->FirstChild( "audio_engine" ) ) ) {
				warningLog( "Error reading configuration file: audio_engine node not found" );
				recreate = true;
			}
			else {
				// Audio driver
				audioDriver = readXmlString( audioEngineNode, "audio_driver", audioDriver );

				// Use metronome
				useMetronome = readXmlBool( audioEngineNode, "use_metronome", useMetronome );

				// Metronome volume
				metronomeVolume = readXmlFloat( audioEngineNode, "metronome_volume", metronomeVolume );

				// Max notes
				maxNotes = readXmlInt( audioEngineNode, "maxNotes", maxNotes );

				//// OSS DRIVER ////
				TiXmlNode* ossDriverNode;
				if ( !(ossDriverNode = audioEngineNode->FirstChild( "oss_driver" ) ) ) {
					warningLog("Error reading configuration file: oss_driver node not found");
					recreate = true;
				}
				else {
					// Buffer size
					bufferSize = readXmlInt( ossDriverNode, "buffer_size", bufferSize );

					// Sample rate
					sampleRate = readXmlInt( ossDriverNode, "samplerate", sampleRate );

					ossDevice = readXmlString( ossDriverNode, "ossDevice", ossDevice );
				}

				//// JACK DRIVER ////
				TiXmlNode* jackDriverNode;
				if ( !(jackDriverNode = audioEngineNode->FirstChild( "jack_driver" ) ) ) {
					warningLog("Error reading configuration file: jack_driver node not found");
					recreate = true;
				}
				else {
					// Jack port name 1
					jackPortName1 = readXmlString( jackDriverNode, "jack_port_name_1", jackPortName1 );

					// Jack port name 2
					jackPortName2 = readXmlString( jackDriverNode, "jack_port_name_2", jackPortName2 );

					// Jack transport slave?
					jackTransportSlave = readXmlBool( jackDriverNode, "jack_transport_slave", jackTransportSlave );

					// Jack track outs?
					jackTrackOuts = readXmlBool( jackDriverNode, "jack_track_outs", jackTrackOuts );

					// Connect to default jack outputs?
					jackConnectDefaults = readXmlBool( jackDriverNode, "jack_connect_defaults", jackConnectDefaults );
				}

				//// ALSA MIDI DRIVER ////
				TiXmlNode* alsaMidiDriverNode;
				if ( !(alsaMidiDriverNode = audioEngineNode->FirstChild( "alsa_midi_driver" ) ) ) {
					warningLog( "Error reading configuration file: alsa_midi_driver node not found" );
					recreate = true;
				}
				else {
					// midi port channel
					midiPortChannel = readXmlInt( alsaMidiDriverNode, "midi_port_channel", midiPortChannel );

					// midi destination name
					midiDest_name = readXmlString( alsaMidiDriverNode, "midi_dest_name", midiDest_name );

					// midi destination client
					midiDest_client = readXmlInt( alsaMidiDriverNode, "midi_dest_client", midiDest_client );

					// midi destination port
					midiDest_port = readXmlInt( alsaMidiDriverNode, "midi_dest_port", midiDest_port );

					// ignore note off
					m_bIgnoreMidiNoteOff = readXmlBool( alsaMidiDriverNode, "ignoreMidiNoteOff", m_bIgnoreMidiNoteOff );
				}
			}

			/////////////// GUI //////////////
			TiXmlNode* guiNode;
			if ( !(guiNode = rootNode->FirstChild( "gui" ) ) ) {
				warningLog("Error reading configuration file: gui node not found");
				recreate = true;
			}
			else {
				// Interface mode
				interfaceMode = readXmlString( guiNode, "interface_mode", interfaceMode );

				// Application font family
				applicationFontFamily = readXmlString( guiNode, "application_font_family", applicationFontFamily );

				// Application font pointSize
				applicationFontPointSize = readXmlInt( guiNode, "application_font_pointsize", applicationFontPointSize );

				// mixer font family
				mixerFontFamily = readXmlString( guiNode, "mixer_font_family", mixerFontFamily );

				// mixer font pointSize
				mixerFontPointSize = readXmlInt( guiNode, "mixer_font_pointsize", mixerFontPointSize );

				// Mixer falloff speed
				mixerFalloffSpeed = readXmlFloat( guiNode, "mixer_falloff_speed", mixerFalloffSpeed );

				// pattern editor grid resolution
				m_nPatternEditorGridResolution = readXmlInt( guiNode, "patternEditorGridResolution", m_nPatternEditorGridResolution );
				m_bPatternEditorUsingTriplets = readXmlBool( guiNode, "patternEditorUsingTriplets", m_bPatternEditorUsingTriplets );
				m_bShowInstrumentPeaks = readXmlBool( guiNode, "showInstrumentPeaks", m_bShowInstrumentPeaks );

				// pattern editor grid height
				m_nPatternEditorGridHeight = readXmlInt( guiNode, "patternEditorGridHeight", m_nPatternEditorGridHeight );

				// mainForm window properties
				setMainFormProperties( readWindowProperties( guiNode, "mainForm_properties", mainFormProperties ) );
				setMixerProperties( readWindowProperties( guiNode, "mixer_properties", mixerProperties ) );
				setPatternEditorProperties( readWindowProperties( guiNode, "patternEditor_properties", patternEditorProperties ) );
				setSongEditorProperties( readWindowProperties( guiNode, "songEditor_properties", songEditorProperties ) );
				setDrumkitManagerProperties( readWindowProperties( guiNode, "drumkitManager_properties", drumkitManagerProperties ) );
				setAudioEngineInfoProperties( readWindowProperties( guiNode, "audioEngineInfo_properties", audioEngineInfoProperties ) );
			}

			/////////////// FILES //////////////
			TiXmlNode* filesNode;
			if ( !(filesNode = rootNode->FirstChild( "files" ) ) ) {
				warningLog("Error reading configuration file: files node not found");
				recreate = true;
			}
			else {
				// last used song
				lastSongFilename = readXmlString( filesNode, "lastSongFilename", lastSongFilename );
			}

		} // rootNode
		else {
			warningLog("Error reading configuration file: hydrogen_preferences node not found");
			recreate = true;
		}
	}
	else {
		warningLog("Configuration file not found.");
		//createPreferencesFile();
		recreate = true;
	}


	// The preferences file should be recreated?
	if (recreate == true) {
		infoLog("Recreating configuration file");
		savePreferences();
	}

}




void PreferencesMng::writeXmlText( TiXmlNode *parent, string name, string text ) {

	TiXmlElement versionNode( name );
	TiXmlText versionText( text );
	versionNode.InsertEndChild( versionText );
	parent->InsertEndChild( versionNode );

}

void PreferencesMng::writeXmlText( TiXmlNode *parent, string name, float value ) {
	char value_char[80];
	sprintf( value_char, "%f", value );
	writeXmlText( parent, name, string(value_char) );

}

void PreferencesMng::writeXmlText( TiXmlNode *parent, string name, int value ) {
	char value_char[80];
	sprintf( value_char, "%d", value );
	writeXmlText( parent, name, string(value_char) );

}



/**
 *
 */
string PreferencesMng::readXmlString( TiXmlNode* parent, string nodeName, string defaultValue ) {
	TiXmlNode* node;
	if ( ( node = parent->FirstChild( nodeName.c_str() ) ) ) {
		if ( node->FirstChild() ) {
			return node->FirstChild()->Value();
		}
		else {
			warningLog( "Using default value in " + nodeName );
			return defaultValue;
		}
	}
	else {
		warningLog( "Error reading XML node: " + nodeName + " node not found" );
		return defaultValue;
	}
}




/**
 *
 */
bool PreferencesMng::readXmlBool( TiXmlNode* parent, string nodeName, bool defaultValue ) {
	TiXmlNode* node;
	if ( ( node = parent->FirstChild( nodeName.c_str() ) ) ) {
		if ( node->FirstChild() ) {
			if ( string( node->FirstChild()->Value() ) == "true" ) {
				return true;
			}
			else {
				return false;
			}
		}
		else {
			warningLog( "Using default value in " + nodeName );
			return defaultValue;
		}
	}
	else {
		warningLog( "Error reading XML node: " + nodeName + " node not found" );
		return defaultValue;
	}
}



/**
 *
 */
float PreferencesMng::readXmlFloat( TiXmlNode* parent, string nodeName, float defaultValue ) {
	TiXmlNode* node;
	if ( ( node = parent->FirstChild( nodeName.c_str() ) ) ) {
		if ( node->FirstChild() ) {
			return atof( node->FirstChild()->Value() );
		}
		else {
			warningLog( "Using default value in " + nodeName );
			return defaultValue;
		}
	}
	else {
		warningLog( "Error reading XML node: " + nodeName + " node not found" );
		return defaultValue;
	}
}



/**
 *
 */
int PreferencesMng::readXmlInt( TiXmlNode* parent, string nodeName, int defaultValue ) {
	TiXmlNode* node;
	if ( ( node = parent->FirstChild( nodeName.c_str() ) ) ) {
		if ( node->FirstChild() ) {
			return atoi( node->FirstChild()->Value() );
		}
		else {
			warningLog( "Using default value in " + nodeName );
			return defaultValue;
		}
	}
	else {
		warningLog( "Error reading XML node: " + nodeName + " node not found" );
		return defaultValue;
	}
}



//----------------------------------------------------------------------------
/**
 * Save the preferences file
 */
void PreferencesMng::savePreferences() {
	infoLog("Saving preferences file");
	string prefDir = getUserHome() + "/.hydrogen";
	string filename = prefDir + "/hydrogen.conf";

	TiXmlDocument doc(filename.c_str());

	TiXmlElement rootNode("hydrogen_preferences");

	// hydrogen version
	writeXmlText( &rootNode, "version", string(VERSION) );

	////// GENERAL ///////
	string restoreLastSongStr = "false";
	if ( restoreLastSong ) {
		restoreLastSongStr = "true";
	}
	writeXmlText( &rootNode, "restoreLastSong", restoreLastSongStr );

	// hear new notes in the pattern editor
	string hearNewNotesStr = "false";
	if ( hearNewNotes ) {
		hearNewNotesStr = "true";
	}
	writeXmlText( &rootNode, "hearNewNotes", hearNewNotesStr );

	//---- AUDIO ENGINE ----
	TiXmlElement audioEngineNode( "audio_engine" );
	{
		// audio driver
		writeXmlText( &audioEngineNode, "audio_driver", audioDriver );

		// use metronome
		string useMetronomeStr = "false";
		if ( useMetronome ) {
			useMetronomeStr = "true";
		}
		writeXmlText( &audioEngineNode, "use_metronome", useMetronomeStr );


		// Metronome volume
		writeXmlText( &audioEngineNode, "metronome_volume", metronomeVolume );

		// Max notes
		writeXmlText( &audioEngineNode, "maxNotes", (int)maxNotes );


		//// OSS DRIVER ////
		TiXmlElement ossDriverNode("oss_driver");
		{
			// buffer size
			writeXmlText( &ossDriverNode, "buffer_size", (int)bufferSize );

			// sample rate
			writeXmlText( &ossDriverNode, "samplerate", (int)sampleRate );

			writeXmlText( &ossDriverNode, "ossDevice", ossDevice );
		}
		audioEngineNode.InsertEndChild( ossDriverNode );

		//// JACK DRIVER ////
		TiXmlElement jackDriverNode( "jack_driver" );
		{
			// jack port name 1
			writeXmlText( &jackDriverNode, "jack_port_name_1", jackPortName1 );
			// jack port name 2
			writeXmlText( &jackDriverNode, "jack_port_name_2", jackPortName2 );

			// jack transport slave
			string jackTransportSlaveString = "false";
			if (jackTransportSlave) {
				jackTransportSlaveString = "true";
			}
			writeXmlText( &jackDriverNode, "jack_transport_slave", jackTransportSlaveString );

			// jack default connection
			string jackConnectDefaultsString = "false";
			if (jackConnectDefaults) {
				jackConnectDefaultsString = "true";
			}
			writeXmlText( &jackDriverNode, "jack_connect_defaults", jackConnectDefaultsString );

			// jack track outs
			string jackTrackOutsString = "false";
			if (jackTrackOuts) {
				jackTrackOutsString = "true";
			}
			writeXmlText( &jackDriverNode, "jack_track_outs", jackTrackOutsString );

		}
		audioEngineNode.InsertEndChild( jackDriverNode );

		//// ALSA MIDI DRIVER ////
		TiXmlElement alsaMidiDriverNode( "alsa_midi_driver" );
		{
			// Midi port channel
			writeXmlText( &alsaMidiDriverNode, "midi_port_channel", midiPortChannel );

			// Midi destination name
			writeXmlText( &alsaMidiDriverNode, "midi_dest_name", midiDest_name );

			// Midi destination client
			writeXmlText( &alsaMidiDriverNode, "midi_dest_client", midiDest_client );

			// Midi destination port
			writeXmlText( &alsaMidiDriverNode, "midi_dest_port", midiDest_port );

			// Ignore midi note off
			string sIgnore = "false";
			if (m_bIgnoreMidiNoteOff) {
				sIgnore = "true";
			}
			writeXmlText( &alsaMidiDriverNode, "ignoreMidiNoteOff", sIgnore );
		}
		audioEngineNode.InsertEndChild( alsaMidiDriverNode );

	}
	rootNode.InsertEndChild( audioEngineNode );

	//---- GUI ----
	TiXmlElement guiNode("gui");
	{
		// Interface mode
		if (interfaceMode == "Top level") {
			writeXmlText( &guiNode, "interface_mode", "Top level" );
		}
		else {
			writeXmlText( &guiNode, "interface_mode", "Child frame" );
		}

		// Application font family
		writeXmlText( &guiNode, "application_font_family", applicationFontFamily );

		// Application font pointsize
		writeXmlText( &guiNode, "application_font_pointsize", applicationFontPointSize );

		// Mixer font family
		writeXmlText( &guiNode, "mixer_font_family", mixerFontFamily );

		// Mixer font pointsize
		writeXmlText( &guiNode, "mixer_font_pointsize", mixerFontPointSize );

		// Mixer falloff speed
		writeXmlText( &guiNode, "mixer_falloff_speed", mixerFalloffSpeed );

		// pattern editor grid resolution
		writeXmlText( &guiNode, "patternEditorGridResolution", m_nPatternEditorGridResolution );

		// pattern editor grid height
		writeXmlText( &guiNode, "patternEditorGridHeight", (int)m_nPatternEditorGridHeight );

		// Pattern editor using triplets?
		string sTripletsString = "false";
		if (m_bPatternEditorUsingTriplets) {
			sTripletsString = "true";
		}
		writeXmlText( &guiNode, "patternEditorUsingTriplets", sTripletsString );

		// show instrument peaks?
		string sInstrPeaksString = "false";
		if (m_bShowInstrumentPeaks) {
			sInstrPeaksString = "true";
		}
		writeXmlText( &guiNode, "showInstrumentPeaks", sInstrPeaksString );

		// MainForm window properties
		writeWindowProperties( &guiNode, "mainForm_properties", mainFormProperties );
		writeWindowProperties( &guiNode, "mixer_properties", mixerProperties );
		writeWindowProperties( &guiNode, "patternEditor_properties", patternEditorProperties );
		writeWindowProperties( &guiNode, "songEditor_properties", songEditorProperties );
		writeWindowProperties( &guiNode, "drumkitManager_properties", drumkitManagerProperties );
		writeWindowProperties( &guiNode, "audioEngineInfo_properties", audioEngineInfoProperties );
	}
	rootNode.InsertEndChild( guiNode );

	//---- FILES ----
	TiXmlElement filesNode( "files" );
	{
		// last used song
		writeXmlText( &filesNode, "lastSongFilename", lastSongFilename );
	}
	rootNode.InsertEndChild( filesNode );

	doc.InsertEndChild(rootNode);
	doc.SaveFile();
}




//----------------------------------------------------------------------------
/**
 * Create preferences directory
 */
void PreferencesMng::createPreferencesDirectory() {
	string prefDir = getUserHome() + "/.hydrogen";

	warningLog("Creating preference file directory in " + prefDir);

	mkdir(prefDir.c_str(),S_IRWXU);

}




//----------------------------------------------------------------------------
/**
 * Create data directory
 */
void PreferencesMng::createDataDirectory() {
	string dir = getUserHome() + "/.hydrogen/data";

	warningLog("Creating data directory in " + dir);

	mkdir(dir.c_str(),S_IRWXU);
}




/**
 * Read the xml nodes related to window properties
 */
WindowProperties PreferencesMng::readWindowProperties( TiXmlNode *parent, string windowName, WindowProperties defaultProp ) {
	WindowProperties prop = defaultProp;

	TiXmlNode* windowPropNode;
	if ( !(windowPropNode = parent->FirstChild( windowName.c_str() ) ) ) {
		warningLog( "Error reading configuration file: " + windowName + " node not found" );
	}
	else {
		// x
		prop.x = readXmlInt( windowPropNode, "x", prop.x );

		// y
		prop.y = readXmlInt( windowPropNode, "y", prop.y );

		// width
		prop.width = readXmlInt( windowPropNode, "width", prop.width );

		// height
		prop.height = readXmlInt( windowPropNode, "height", prop.height );
	}

	return prop;
}






/**
 * Write the xml nodes related to window properties
 */
void PreferencesMng::writeWindowProperties(TiXmlNode *parent, string windowName, WindowProperties prop ) {
	TiXmlElement windowPropNode( windowName.c_str() );
		writeXmlText( &windowPropNode, "x", prop.x );
		writeXmlText( &windowPropNode, "y", prop.y );
		writeXmlText( &windowPropNode, "width", prop.width );
		writeXmlText( &windowPropNode, "height", prop.height );
	parent->InsertEndChild( windowPropNode );
}




/**
 * Constructor
 */
WindowProperties::WindowProperties() {
//	infoLog( "INIT" );
	x = 0;
	y = 0;
	width = 0;
	height = 0;
}



/**
 * Destructor
 */
WindowProperties::~WindowProperties() {
//	infoLog( "DESTROY" );
}


